#  Artifact Source client


from typing import Optional, List
from pydantic import BaseModel, Field
import enum

from .jiggy_session import session
from ..user import Team, authenticated_user, all_teams

class ArtifactService(str, enum.Enum):
    """
    The set of backend artifact stores that we know how to access
    """    
    wandb   = 'Weights & Biases'
    hf      = 'Hugging Face'
    jb      = 'Jiggyboard'


class ArtifactSource(BaseModel):
    """
    The sources of the models, weights, and (eval) code that we deal with
    """
    id:         int             = Field(description='Unique Configuration ID')
    team_id:    int             = Field(description='The team_id that owns this artifact source')
    source:     ArtifactService = Field(description='Source of Artifacts such as datasets and models')
    api_key:    str             = Field(description='API Key to access the datasource service')
    host:       Optional[str]   = Field(default=None, description='alterative (e.g. private) hostname')
    created_by: int             = Field(description='The user_id that created this item.')
    updated_by: int             = Field(description='The user_id that last modified this item.')

    def delete(self):
        """
        attempt to delete this ArtifactSource
        """
        session.delete(f"/teams/{self.team_id}/artifact_sources/{self.id}")
                       
    
class ArtifactSourceRequest(BaseModel):
    """
    Create an ArtifactSource
    """
    source:  ArtifactService = Field(description='Source of Artifacts such as datasets and models')
    api_key: str             = Field(description='API Key to access the datasource service')
    host:    Optional[str]   = Field(default=None, description='alterative (e.g. private) hostname')



def artifact_sources(team : Team = None) -> List[ArtifactSource]:
    """
    return list of all artifact sources
    """
    if team:
        r = session.get(f"/teams/{team.id}/artifact_sources")
        items = r.json()['items']
    else:
        items = []
        for team in all_teams():
            r = session.get(f"/teams/{team.id}/artifact_sources")
            items.extend(r.json()['items'])
    return [ArtifactSource(**i) for i in items]





# Jiggy Client

import os
from typing import Optional, List
from pydantic import BaseModel, Field
import enum

from .jiggy_session import JiggySession


session = JiggySession('jiggyuser-v0')
    

###
## API Key
###


class  ApiKey(BaseModel):
    key:         str           = Field(default=None, description='The api key.')
    description: Optional[str] = Field(default=None, description='Optional user supplied description of the key.')
    created_at:  float         = Field(description='The epoch timestamp when the key was created.')
    last_used :  float         = Field(description='The epoch timestamp when the key was last used to create a JWT.')
    
###
## User
###

class User(BaseModel):
    id:              int = Field(description='Internal user_id')
    username:        str = Field(min_length=3, max_length=39, description='Unique name for the user.')
    auth0_userid:    str = Field(description='Auth0 userid.  This can be None for anonymous accounts created via api key')
    default_team_id: int = Field(description='The default team for this user')


class UserPostRequest(BaseModel):
    username:        str = Field(min_length=3, max_length=39, description='Unique name for the user.')
    

class UserPostPatchRequest(BaseModel):
    default_team_id: Optional[int] = Field(description='The default team for this user')
    username:        Optional[str] = Field(min_length=3, max_length=39, description='Unique name for the user.')

    
###
## Team
###

class Team(BaseModel):
    id:          int           = Field(description='Internal team id')
    name:        str           = Field(min_length=3, max_length=39, description='Unique name for this team.')
    description: Optional[str] = Field(default=None, description='Optional user supplied description.')
    created_at:  float         = Field(description='The epoch timestamp when the team was created.')
    updated_at:  float         = Field(description='The epoch timestamp when the team was updated.')


class TeamRole(str, enum.Enum):
    admin   = 'admin'
    member  = 'member'
    service = 'service'
    view    = 'view'
    

class TeamPostRequest(BaseModel):
    name:        str           = Field(min_length=3, max_length=39, description='Unique name for this team.')
    description: Optional[str] = Field(default=None, description='Optional user supplied description.')

class TeamMemberPostRequest(BaseModel):
    username:    str           = Field(description='The user_id of a member to invite to the team.')
    role:        TeamRole      = Field(description='The users role in the team')

class TeamMember(BaseModel):
    id:                  int      = Field(description="Unique membership id")
    username:            str      = Field(description="Member username")
    created_at:          float    = Field(description='The epoch timestamp when the membership was created.')
    updated_at:          float    = Field(description='The epoch timestamp when the membership was updated.')
    invited_by_username: str      = Field(description="The username that invited this member to the team.")
    role:                TeamRole = Field(description="The user's role in the team")
    accepted:            bool     = Field(description='True if the user has accepted the team membership.')
    
        
###
##  Teams
###
class Team(BaseModel):
    id:          int           = Field(description='Internal team id')
    name:        str           = Field(min_length=3, max_length=39, description='Unique name for this team.')
    description: Optional[str] = Field(default=None, description='Optional user supplied description.')
    created_at:  float         = Field(description='The epoch timestamp when the team was created.')
    updated_at:  float         = Field(description='The epoch timestamp when the team was updated.')

    def members(self) -> List[TeamMember]:
        return [TeamMember(**tm) for tm in session.get(f'/teams/{self.id}/members').json()['items']]
    
    def add_member(self, username : str, role : TeamRole) -> TeamMember:
        """
        invite the specified username to this team with the corresponding TeamRole
        """
        rsp = session.post(f"/teams/{self.id}/members", json={"username": username, 'role':role})
        return TeamMember(**rsp.json())

    def delete_member(self, username : str):
        """
        attempt to remove the specified username from this team
        """
        member = [m for m in self.members() if m.username == username]
        if not member:
            raise Exception(f"{username} not found in Team")
        session.delete(f"/teams/{self.id}/members/{member[0].id}")

    def set_default(self):
        """
        Set this team as the user's default team.
        """
        user_id = authenticated_user().id
        session.patch(f"/users/{user_id}", json={'default_team_id': self.id})
        

def all_teams() -> List[Team]:
    """
    return all Teams that the user is a member of
    """
    resp = session.get('/teams')
    return [Team(**i) for i in resp.json()['items']]


def team(name_or_id) -> Team:
    """
    get team by name or id
    raises Exception if an exact match for name is not found
    """
    teams = [t for t in session.get('/teams').json()['items'] if t['name'] == name_or_id or t['id'] == name_or_id]
    if len(teams):
        return Team(**teams[0])
    raise Exception(f'Team "{name}" not found')


def api_keys() -> List[ApiKey]:
    return [ApiKey(**i) for i in session.get('/apikey').json()['items']]


def authenticated_user() -> User:
    """
    return the authenticated user's User object
    """
    return User(**session.get("/users/current").json())


def create_team(name : str) -> Team:
    resp = session.post("/teams", json={"name":name})
    return Team(**resp.json())






#  Board and related client


from typing import Optional, List
from pydantic import BaseModel, Field
from pydantic import condecimal
import enum

from .jiggy_session import session

from .artifact_source import ArtifactSource
from ..user import Team, authenticated_user, all_teams



###
##   Evaluation
###

class EvalStatus(str, enum.Enum):
    """
    The status of an evaluation.
    """
    scheduled = 'scheduled'
    running   = 'running'
    failed    = 'failed'    # failed to execute evaluation code due to runtime error
    error     = 'error'     # error occured in the evaluation code
    complete  = 'complete'


class EvaluationPatchRequest(BaseModel):
    """
    An evaluation of a specific version of code, model, and dataset.
    """
    status:           EvalStatus    = Field(index=True, description='The status of the evaluation')
    percent_complete: Optional[int] = Field(default=0, ge=0, le=100, description='The evaluation progress as an integer completeness percentage.')
    

class Evaluation(BaseModel):
    """
    An evaluation of a specific version of code, model, and dataset.
    """
    id:               int           = Field(description='Unique eval ID')
    team_id:          int           = Field(description='The Team with which this artifact series is associated.')
    board_id:         int           = Field(description='The Board with which this Evaluation is associated.')
    code_id:          int           = Field(description='The eval code used for the Evaluation')
    model_id:         int           = Field(description='The model used for the Evaluation')
    dataset_id:       int           = Field(description='The dataset used for the Evaluation')
    env_id:           int           = Field(description='The environment used for the Evaluation')
    status:           EvalStatus    = Field(description='The status of the evaluation')
    created_at:       float         = Field(description='The epoch timestamp when the Evaluation was created.')
    updated_at:       float         = Field(description='The epoch timestamp when the Evaluation was updated.')
    ipynb_output:     Optional[str] = Field(default=None, description='The uri of the ipynb output of the evaluation run')
    html_output:      Optional[str] = Field(default=None, description='The uri of the html output of the evaluation run')
    pdf_output:       Optional[str] = Field(default=None, description='The uri of the pdf output of the evaluation run')
    percent_complete: Optional[int] = Field(default=0, ge=0, le=100, description='The evaluation progress as an integer completeness percentage.')

    def update_status(status:           EvalStatus,
                      percent_complete: int = 0):
        model = EvaluationPatchRequest(**locals())
        rsp = session.patch(f'/teams/{self.team_id}/boards/{self.board_id}/evaluations/{self.id}', model=model)
        for key, value in rsp.json().dict():
            setattr(self, key, value)

    

    
###
##   Artifact
###

class ArtifactType(str, enum.Enum):
    """
    the types of artifact we deal in
    """
    code    = 'code'     
    model   = 'model'
    dataset = 'dataset'



    
class Artifact(BaseModel):
    """
    A specific version of an Artifact that we have discovered
    """
    id:          int          = Field(description='Unique ID of a specific version of a BoardArtifact')
    team_id:     int          = Field(description='The team that owns the board.')
    board_id:    int          = Field(description='The Board with which this artifact is associated.')
    name:        str          = Field(description="The full name of the artifact including version info.")
    jiggy_type:  ArtifactType = Field(description='The type of this artifact in jiggy: "code", "dataset", or "model".')    
    created_at:  float        = Field(description='The epoch timestamp when the ArtifactVersion was created in Jiggyboard.')


###
##    ArtifactWatcher 
###


class WatcherDefault(str, enum.Enum):
    """
    Configures an ArtifactWatcher to default to processing all new versions as they are discovered or ignoring new versions.
    
    process:  process (evaluate) all new artifact versions as they are discovered.  
              This can be modified by additional WatcherRules to exclude certain versions/aliases from processing.

    ignore:   ignore all new artifact versions as they are discovered.  
              This can be modified by additional WatcherRules which enable processing for certain versions/aliases.
    """
    process  = 'process'
    ignore   = 'ignore'

    
class ArtifactWatcher(BaseModel):
    """
    An Artifact that a board is configured to use
    """
    id:           int            = Field(description='Unique ArtifactWatcher ID.')
    team_id:      int            = Field(description='The team that owns the board.')
    board_id:     int            = Field(description='The Board with which this artifact is associated.')
    name:         str            = Field(description='A versionless artifact name in the form of "entity/project/name", as known to the backend service')
    jiggy_type:   ArtifactType   = Field(description='The type of this artifact in jiggy: "code", "dataset", or "model".')
    source_id:    int            = Field(description='The backend service configuration to access artifact')
    source_type:  str            = Field(description='The type of the artifact as known to the backend service.')
    enabled:      bool           = Field(description='True if evaluation is enabled for this artifact, False if disabled.')    
    default:      WatcherDefault = Field(description="Sets the default policy to process or ignore new artifact versions. Modified by WatcherRules.")
    created_at:   float          = Field(description='The epoch timestamp when the Artifact was created.')
    updated_at:   float          = Field(description='The epoch timestamp when the Artifact was updated.')
    created_by:   int            = Field(description='The user_id that created this item.')
    updated_by:   int            = Field(description='The user_id that last modified this item.')

    def delete(self):
        session.delete(f'/teams/{self.team_id}/boards/{self.board_id}/watchers/{self.id}')



class ArtifactWatcherPostRequest(BaseModel):
    """
    Create a ArtifactWatcher
    """
    source_id:    int            = Field(description='The configuration to access artifact')
    jiggy_type:   ArtifactType   = Field(description='The type of this artifact.')
    name:         str            = Field(description='The name of the artifact as found in the Artifact Source.')
    enabled:      Optional[bool] = Field(default=True, description='True if evaluation is enabled for this artifact, False if disabled.')
    default_all:  Optional[bool] = Field(default=True, description="True to process all versions by default. False to process no versions by default. Policy modified by ArtifactWatcherRules")


class ArtifactWatcherPatchRequest(BaseModel):
    """
    Modify a ArtifactWatcher.
    """
    enabled:       Optional[bool] = Field(default=None, description='True if evaluation is enabled for this artifact, False if disabled.')
    default_all:   Optional[bool] = Field(default=None, description="True to process all versions by default. False to process no versions by default. Policy modified by ArtifactWatcherRules")
    


    
###
##  Board
###
    
class Board(BaseModel):
    """
    A very Jiggy leaderboard that tracks results across a range of different model, dataset versions, and eval code versions
    """
    id:          int    = Field(description='Unique Board ID')
    name:        str    = Field(description='The name of the Board.  Must be unique in team context.')
    team_id:     int    = Field(description='The team that owns this Board.')
    public:      bool   = Field(description='True if Board is publicly visible, False if private to a team.')
    enabled:     bool   = Field(description='True if evaluation is enabled for this Board, False if disabled.')
    created_at:  float  = Field(description='The epoch timestamp when the Board was created.')
    updated_at:  float  = Field(description='The epoch timestamp when the Board was updated.')
    created_by:  int    = Field(description='The user_id that created this item.')
    updated_by:  int    = Field(description='The user_id that last modified this item.')

    def delete(self):
        session.delete(f'/teams/{self.team_id}/boards/{self.id}')

    def create_artifact_watcher(self,
                                source:      ArtifactSource,
                                jiggy_type:  ArtifactType,
                                name:        str,
                                enabled:     bool = True,
                                default_all: bool = True) -> ArtifactWatcher:

        post_model = ArtifactWatcherPostRequest(board_id = self.id,
                                                source_id = source.id,
                                                **locals())
        
        rsp = session.post(f'/teams/{self.team_id}/boards/{self.id}/watchers', model = post_model)
        return ArtifactWatcher(**rsp.json())
                           

    def artifact_watchers(self) -> List[ArtifactWatcher]:
        rsp = session.get(f"/teams/{self.team_id}/boards/{self.id}/watchers")
        return [ArtifactWatcher(**w) for w in rsp.json()['items']]
    

    def artifacts(self) -> List[Artifact]:
        rsp = session.get(f'/teams/{self.team_id}/boards/{self.id}/artifacts')
        return [Artifact(**a) for a in rsp.json()['items']]

    def artifact(self, artifact_id) -> Artifact:
        rsp = session.get(f'/teams/{self.team_id}/boards/{self.id}/artifacts/{artifact_id}')
        return Artifact(**rsp.json())
    
    def evaluations(self, status : EvalStatus = None) -> List[Evaluation]:
        if status:
            rsp = session.get(f'/teams/{self.team_id}/boards/{self.id}/evaluations?status={status}')
        else:
            rsp = session.get(f'/teams/{self.team_id}/boards/{self.id}/evaluations')
        return [Evaluation(**e) for e in rsp.json()['items']]


    def evaluation(self, eval_id : int) -> Evaluation:
        rsp = session.get(f'/teams/{self.team_id}/boards/{self.id}/evaluations/{eval_id}')
        return Evaluation(**rsp.json())

    
class BoardPostRequest(BaseModel):
    """
    Create a Board item
    """
    name:        str             = Field(description='The name of the Board.  Must be unique in team context.')
    description: Optional[str]   = Field(description='The description of this board.')    
    public:      Optional[bool]  = Field(default=False, description='True if Board is publicly visible, False if private to a team.')
    enabled:     Optional[bool]  = Field(default=True, description='True if evaluation is enabled for this Board, False if disabled.')


class BoardPatchRequest(BaseModel):
    """
    Modify a Board item
    """
    name:        Optional[str]   = Field(default=None,  description='The name of the Board.  Must be unique in team context.')
    public:      Optional[bool]  = Field(default=None,  description='True if Board is publicly visible, False if private to a team.')
    enabled:     Optional[bool]  = Field(default=None,  description='True if evaluation is enabled for this Board, False if disabled.')
    
    



    
###
##    WatcherRule
### 

class WatcherRule(BaseModel):
    """
    A rule applied to a ArtifactWatcher that serves to configure or limit evaluation of specific versions or aliases.
    """
    id:             int           = Field(description='Unique ID')
    artifact_id:    int           = Field(description='The ArtifactWatcher this Rule is associated with')
    alias:          Optional[str] = Field(description='Configure evaluation for this alias of the artifact.')
    ignore_alias:   Optional[str] = Field(description='Ignore evaluation for this alias of the artifact.')
    version:        Optional[str] = Field(description='Configure evaluation for this specific version of the artifact.')
    ignore_version: Optional[str] = Field(description='Disable evaluation for this specific version of the artifact.')
    created_at:     float         = Field(description='The epoch timestamp when the Artifact was created.')
    updated_at:     float         = Field(description='The epoch timestamp when the Artifact was updated.')
    created_by:     int           = Field(description='The user_id that created this item.')
    updated_by:     int           = Field(description='The user_id that last modified this item.')



class WatcherRulePostRequest(BaseModel):
    """
    Used to create a WatcherRule
    """
    artifact_id:    int           = Field(description='The ArtifactWatcher this Rule is associated with')
    alias:          Optional[str] = Field(description='Configure evaluation for this alias of the artifact.')
    ignore_alias:   Optional[str] = Field(description='Ignore evaluation for this alias of the artifact.')
    version:        Optional[str] = Field(description='Configure evaluation for this specific version of the artifact.')
    ignore_version: Optional[str] = Field(description='Disable evaluation for this specific version of the artifact.')



def create_board(name:        str,
                 team:        Team = None,
                 description: str  = None,
                 public:      bool = False,
                 enabled:     bool = True) -> Board:
    """
    Create a board with the specified parameters
    
    If no team is specified, the user's default team will be used
    """
    if team is None:
        team_id = authenticated_user().default_team_id
    else:
        team_id = Team.id
    rsp = session.post(f"/teams/{team_id}/boards", model=BoardPostRequest(**locals()))
    return Board(**rsp.json())


    
def boards(team : Team = None) -> List[Board]:
    """
    Get the boards for the specified team.
    If no team is specified, the boards for all of the user's teams will be returned.
    """
    if team:
        r = session.get(f"/teams/{team.id}/boards")
        items = r.json()['items']
    else:
        items = []
        for team in all_teams():
            r = session.get(f"/teams/{team.id}/boards")
            items.extend(r.json()['items'])    
    return [Board(**b) for b in items]
    



def get_board(name : str,
          team : Team = None) -> Board:
    """
    return a board of the specified name.
    If team is None then return the board from any of the teams.  
    Note: a board name could be duplicated across multiple teams. 
    In this case this function returns the first matching name it finds.
    """
    for board in boards(team):
        if board.name == name:
            return board
    



#
# Jiggyboard interface to wandb
#

import wandb

from ..login import window_open
from ..user import authenticated_user, all_teams, Team

from .artifact_source import ArtifactService, ArtifactSource, ArtifactSourceRequest

from .jiggy_session import session




def create_artifact_source(team    : Team = None,
                           api_key : str  = None,
                           host    : str  = "api.wandb.ai") -> ArtifactSource:
    """
    Create a Weights & Biases ArtifactSource

    If team is None then will use user's default team (as specified in User.default_team_id
    api_key is the wandb api key to use to access the artifacts stored on wandb. 
    host is the wandb api hostname, defaulting to api.wandb.ai
    """
    if team is None:
        # get user default team
        # not sure this is a good idea, may want to require team
        team_id = authenticated_user().default_team_id
    else:
        team_id = Team.id

    # prompt user for wandb key
    window_open("https://%s/authorize" % host.split('.',1)[1])
    api_key = input("Enter your wandb.ai API Key: ")
    # api key validation will occur server side
    rsp = session.post(f'/teams/{team_id}/artifact_sources', model=ArtifactSourceRequest(host    = host,
                                                                                         api_key = api_key,
                                                                                         source  = ArtifactService.wandb))
    return ArtifactSource(**rsp.json())



# Jiggy ANN Client

import os
import requests
import urllib.parse

from typing import Optional, List
from pydantic import BaseModel, Field
from time import time, sleep
import enum

import hnswlib


from .jiggy_session import JiggySession


session = JiggySession('jiggy-v0')
#session = JiggySession('jiggyann-v0')


class IndexError(Exception):
    """
    Index creation failed
    """


    
class IndexLibraries(str, enum.Enum):
    """
    The library used to create the index.
    Currently only 'hnswlib' is supported.
    """
    hnswlib = 'hnswlib'


class DistanceMetric(str, enum.Enum):
    """
    The distance metric to use for the nearest neighbor index.
    """
    cosine = 'cosine'
    ip     = 'ip'
    l2     = 'l2'

class IndexBuildState(str, enum.Enum):
    prep = "preparing data"
    indexing = "indexing vectors"
    saving = "saving index"
    complete = "index complete"
    failed = "indexing failure"

    
class Collection(BaseModel):
    
    id:    int = Field(description='Unique identifier for a particular collection.')
    name:  str = Field(description="The Collection's unique name within the team context.")

    dimension: int         = Field(default= 0, description="The dimension of the vectors in this collection.")
    team_id:   int         = Field(None, index=True, description="The team that this collection is associated with.")
    count: int             = Field(default=0, description="The number of vectors in the collection")
    created_at: float = Field(description='The epoch timestamp when the collection was created.')
    updated_at: float = Field(description='The epoch timestamp when the collection was updated.')
       
    def add(self, vector_data:[float], vector_id:int):
        """
        Add the specified vector_data to the collection and associate it with the specified vector_id.
        A collection may not have more than one vector with a given vector_id.
        Re-using a vector_id will over-write any existing  vector_id +  vector pair.
        All vectors in a collection must share the same dimension.
        """
        vector_data = [float(x) for x in vector_data]
        session.post(f"/collections/{self.id}/vectors/{vector_id}", json={'vector': vector_data})
        if not self.dimension:
            self.dimension = len(vector_data)  # record dimension

    def get(self, vector_id : int):
        """
        Return the vector with the specified vector_id from the collection
        """
        resp = session.get(f"/collections/{self.id}/vectors/{vector_id}")
        return resp.json()['vector']
        
    def delete(self, vector_id : int):
        """
        Delete the vector with the specified vector_id from the collection
        """        
        session.delete(f"/collections/{self.id}/vectors/{vector_id}")
        
    def create_index(self,
                     tag='latest',
                     target_library=IndexLibraries.hnswlib,
                     metric=DistanceMetric.cosine,
                     target_recall=None,
                     M=None,
                     ef=None):
        body = {'tag': tag,
                'target_library': target_library,
                'metric': metric,
                'target_recall': target_recall,
                'hnswlib_M':  M,
                'hnswlib_ef': ef}
        resp = session.post("/collections/%d/index" % self.id, json=body)
        return Index(**resp.json(), collection=self)
                                                            
    def delete_collection(self):
        """
        Delete the collection.
        """
        resp = session.delete("/collections/%s" % self.id)
        self.id=None
        self.name= None
        self.created_at=None
        self.updated_at=None
        self.team_id=None
        self.dimension=None


    def get_index(self, tag:str="latest"):
        data = {'tag': tag}
        resp = session.get(f"/collections/{self.id}/index?{urllib.parse.urlencode(data)}")
        items = resp.json()['items']
        if tag and items:
            return Index(**items[0], collection=self)
        return [Index(**i, collection=self) for i in items]

    def get_all_index(self):
        return self.get_index(None)

    def _update(self):
        resp = session.get(f"/collections/{self.id}")
        self.__dict__.update(resp.json())
        
    def __getattribute__(self, attr):
        if attr in ['count', 'updated_at']:
            self._update()
        return object.__getattribute__(self, attr)        

        
def create_collection(name:str, team_name=None):
    """
    Create a collection of the specified name
    Returns a Collection object.
    Raises ClientError if the specified name already exists

    'team_obj_or_name' is optional and can be either a team name or team object.
    if team is unspecified the user's default team will be used.
    """
    data = {'name': name}
    if team_name:
        data['team_id'] = team(team_name).id
    resp = session.post("/collections", json=data)
    return Collection(**resp.json())


def collection(name:str):
    """
    Get a collection by name.
    Returns a Collection Object, or None if collection name is not found
    """
    data = {'name': name}
    resp = session.get(f"/collections?{urllib.parse.urlencode(data)}")
    items = resp.json()['items']
    if len(items):
        return Collection(**items[0])
    raise Exception(f"Collection {name} not found")


def all_collections():
    resp = session.get("/collections")
    return [Collection(**i) for i in resp.json()['items']]


class Index(BaseModel):
    id: int            = Field(description='Unique identifier for a given index.')
    collection: Collection = Field(description="The containing collection")
    tag:  str = Field(description="User tag for this Index.  Uniquely identifies an index in the context of a collection.")
    name: str = Field(description="The full name for this index in the form of TEAM_NAME/COLLECTION_NAME:TAG")
    target_library:    IndexLibraries = Field(default=IndexLibraries.hnswlib, description="The library use to create the index")
    metric:        str = Field(description='The distance metric ("space" in hnswlib): "cosine", "ip", or "l2"')
    hnswlib_M:  Optional[int] = Field(default=None, description="The M value passed to hnswlib when creating the index.")
    hnswlib_ef: Optional[int]  = Field(default=None, description="The ef_construction value passed to hnswlib when creating the index.")
    hnswlib_ef_search: Optional[int] = Field(default=None, ge=10, description="The recommended ef value to use at search time.")
    target_recall: Optional[float] = Field(default=None, ge=0.5, le=1, description="The desired recall value to target for index parameter optimization.")        
    count: int = Field(description="The number of vectors included in the index.  The number of vectors in the collection at the time of index build.")
    created_at: float = Field(description='The epoch timestamp when the index was requested to be created.')
    state: IndexBuildState = Field(description = "The current build status.")
    completed_at: float = Field(description='The epoch timestamp when the index build was completed.')
    build_status: str     = Field(description='Informational status message for the index build.')
    url: Optional[str] = Field(default=None, description='The url the index can be downloaded from.  The url is valid for a limited time.')

    
    def _wait_complete(self):
        while self.state not in [IndexBuildState.failed, IndexBuildState.complete]:
            print(f"Index Status: {self.build_status}")  # XXX provide percentage completion
            data = {'tag': self.tag}
            resp = session.get(f"/collections/{self.collection.id}/index?{urllib.parse.urlencode(data)}")
            items = resp.json()['items'][0]
            self.__dict__.update(items)
            sleep(2)  # XXX update sleep time based on something like halfway toward the estimated remaining completed_at estimate
        if self.state == IndexBuildState.failed:
            raise IndexError(self.build_status)
        
            
    def hnswlib_index(self, max_elements:int = None):
        self._wait_complete()
        if not max_elements:
            max_elements = self.count
        resp = requests.get(self.url)
        assert(resp.status_code == 200)
        assert(int(len(resp.content)) == int(resp.headers['Content-Length']))
        fname = self.name.replace('/','-')
        fname += f".{self.target_library}"
        open(fname, 'wb').write(resp.content)
        ix = hnswlib.Index(space=self.metric, dim=self.collection.dimension)
        ix.load_index(fname, max_elements=max_elements)
        ef = self.hnswlib_ef_search if self.hnswlib_ef_search else self.hnswlib_ef
        ix.set_ef(ef)
        return ix

    def tests(self):
        """
        Return the IndexTest results for this index
        """
        tests = session.get(f'/collections/{self.collection.id}/index/{self.id}/tests').json()['items']
        return [IndexTest(**test, index=self) for test in tests]


class IndexTest(BaseModel):
    id:                   int = Field(description="Unique database identifier for a given index test.")
    index:              Index = Field(description="The index under test.")
    test_count:           int = Field(description="The number of vectors in the test sample.")
    test_k:               int = Field(description="The number of nearest neighbors (k) to query for each test vector.")
    recall:             float = Field(description="The recall of the index based on a comparison to exhaustive KNN.")
    qps:                float = Field(description="The estimated queries per second of the index for vector search with batchsize of 1 (a single vector at a time).")
    cpu_info:             str = Field(description="The CPU that executed the test.")
    created_at:         float = Field(description="The epoch timestamp when the test was completed.")
    hnswlib_ef: Optional[int] = Field(default=None, description="The ef value passed to hnsw_index.set_ef() for this test.")
        
    def __repr__(self):
        return f"{self.index.name:20s}: recall: {self.recall:5.3f}  QPS: {self.qps:7.1f} @ EF={self.hnswlib_ef}"
    

